/*
SGPC: Simple Genetic Programming in C
(c) 1993 by Walter Alden Tackett and Aviram Carmi
 
 This code and documentation is copyrighted and is not in the public domain.
 All rights reserved. 
 
 - This notice may not be removed or altered.
 
 - You may not try to make money by distributing the package or by using the
   process that the code creates.
 
 - You may not distribute modified versions without clearly documenting your
   changes and notifying the principal author.
 
 - The origin of this software must not be misrepresented, either by
   explicit claim or by omission.  Since few users ever read sources,
   credits must appear in the documentation.
 
 - Altered versions must be plainly marked as such, and must not be
   misrepresented as being the original software.  Since few users ever read
   sources, credits must appear in the documentation.
 
 - The authors are not responsible for the consequences of use of this 
   software, no matter how awful, even if they arise from flaws in it.
 
If you make changes to the code, or have suggestions for changes,
let us know!  (gpc@ipld01.hac.com)
*/

#ifndef lint
static char crossover_c_rcsid[]="$Id: crossover.cpp,v 1.3 2006/09/02 22:33:04 yanase Exp $";
#endif

/*
 *
 * $Log: crossover.cpp,v $
 * Revision 1.3  2006/09/02 22:33:04  yanase
 * add rename_tree_hold_type() in treegen.cpp/proto.h
 * to maintain the function node type after crossover and mutation.
 * replaced rename_tree by rename_tree_hold_type().
 * see also crossover.cpp and mutation.cpp.
 *
 * Revision 1.2  2006/09/01 08:59:46  caranha
 * Debugging
 *
 * Revision 1.1.1.1  2006/08/31 08:56:26  yanase
 * This is the implementation of Inductive GP written in Adaptive Learning of Polynomial Networks.
 *
 * Revision 1.1.1.1  2006/08/21 09:52:44  caranha
 * Uploaded strogonoff Unix 02/2006 version to the server
 *
 * Revision 1.4  2006/01/26 15:55:13  caranha
 * code cleanup, output() bug fixed (i think)
 *
 * Revision 1.3  2006/01/25 11:18:11  caranha
 * Added class for parameter manipulation. Moved some globals around.
 *
 * Revision 1.2  2006/01/24 07:45:51  caranha
 * Nothing really important, just moving globals around
 *
 * Revision 1.1  2006/01/24 03:54:24  caranha
 * Added initial files - stroglib is the library, debug-app is a debug application for the library - should not be needed
 *
 * Revision 2.6  1993/04/22  07:39:12  gpc-avc
 * Removed old log messages
 *
 * Revision 2.5  1993/04/14  05:07:57  gpc-avc
 * removed #ifdef TRACE code
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "magic.h"
#include "gpc.h"
#include "proto.h"
#include "problem.h"
#include "getparams.h"
#include "globals.h"

/*
#define TRACEcaap
#define TRACEcafp
*/

VOID crossover_at_any_pt(
  pop_struct 	*pop,
  tree 		*p1,
  tree 		*p2,
  tree 		**o1,
  tree 		**o2,
  int           *jcross1,
  int           *jcross2
  )
{
  tree	*st1, *st2, **st1ptr, **st2ptr, *parent;
  int	xpt1, xpt2;
  int flag1,flag2;

  ncross_at_any_pt = ncross_at_any_pt +1;

  xpt1 = random_int(count_crossover_pts(p1));
  xpt2 = random_int(count_crossover_pts(p2));

  *jcross1 = xpt1;
  *jcross2 = xpt2;

  *o1 = copy_tree(p1);
  *o2 = copy_tree(p2);

  st1 = get_subtree(*o1, xpt1);
  st2 = get_subtree(*o2, xpt2);

  st1ptr = pointer_to_subtree(o1, st1);
  st2ptr = pointer_to_subtree(o2, st2);

  *st1ptr = st2;
  *st2ptr = st1;

  validate_crossover(pop, p1, p2, o1, o2, &flag1, &flag2);

  if (flag1) {
    parent = get_parent(*o1, st2);
    st2->parent = parent;
    //    rename_tree(st2);
    rename_tree_hold_type(st2);
  }
  if (flag2) {
    parent = get_parent(*o2, st1);
    st1->parent = parent;
    //    rename_tree(st1);
    rename_tree_hold_type(st1);
  }

}

void crossover_at_func_pt(
  pop_struct 	*pop,
  tree 		*p1,
  tree 		*p2,
  tree 		**o1,
  tree 		**o2,
  int           *jcross1,
  int           *jcross2
  )
{
  tree	*st1, *st2, **st1ptr, **st2ptr, *parent;
  int	xpt1, xpt2;
  int flag1,flag2;

  ncross_at_funct_pt = ncross_at_funct_pt + 1;

  xpt1 = random_int(count_function_pts(p1));
  xpt2 = random_int(count_function_pts(p2));

  *jcross1 = xpt1;
  *jcross2 = xpt2;

  *o1 = copy_tree(p1);
  *o2 = copy_tree(p2);

  st1 = get_function_subtree(*o1, xpt1);
  st2 = get_function_subtree(*o2, xpt2);

  st1ptr = pointer_to_subtree(o1, st1);
  st2ptr = pointer_to_subtree(o2, st2);

  *st1ptr = st2;
  *st2ptr = st1;

  validate_crossover(pop, p1, p2, o1, o2, &flag1, &flag2);

  if (flag1) {
    parent = get_parent(*o1, st2);
    st2->parent = parent;
    //    rename_tree(st2);
    rename_tree_hold_type(st2);
  }
  if (flag2) {
    parent = get_parent(*o2, st1);
    st1->parent = parent;
    //    rename_tree(st1);
    rename_tree_hold_type(st1);
  }

}


void validate_crossover(
  pop_struct 	*pop,
  tree 		*p1,
  tree 		*p2,
  tree 		**o1,
  tree 		**o2,
  int           *flag1,
  int           *flag2
  )
{
  int	d1, d2;

  d1 = depth_of_tree(*o1);
  d2 = depth_of_tree(*o2);
  *flag1 = 1;
  *flag2 = 1;

  if (d1 > pop[(*o1)->pop].max_depth_after_crossover) {
    free_tree(*o1);
    *o1 = copy_tree(p1);
    *flag1 = 0;

  }
  if (d2 > pop[(*o2)->pop].max_depth_after_crossover) {
    free_tree(*o2);
    *o2 = copy_tree(p2);
    *flag2 = 0;

  }
}

int depth_of_tree(
	tree *t
	)
{
  int	i, n, maxn=0;

  if (t->nodetype == FUNCTION) {
    for (i=0; i<function_arity(t); i++) {
      n = depth_of_tree(t->type.func->arg[i]);
      maxn = max(maxn,n);
    }
    return 1+maxn;
  } else {
    return 0;
  }
}

int count_crossover_pts(
	tree *t
	)
{
  int	i, n=1;

  if (t->nodetype == FUNCTION) {
    for (i=0; i<function_arity(t); i++) {
      n += count_crossover_pts(t->type.func->arg[i]);
    }
  }
  return n;
}

int count_function_pts(
	tree *t
	)
{
  int	i, n=1;

  if (t->nodetype == FUNCTION) {
    for (i=0; i<function_arity(t); i++) {
      n += count_function_pts(t->type.func->arg[i]);
    }
    return n;
  } else {
    return 0;
  }
}

tree *get_subtree(
	tree *t,
	int n
	)
{
  int	m;

  m = n;
  return gs(t,&m);
}

tree *gs(
	tree *t,
	int *n
	)
{
  tree	*st;
  int	i;
 
  if (!(*n)) return(t);
  else if (t->nodetype == FUNCTION) {
    for (i=0; i<function_arity(t); i++) {
      --(*n);
      st = gs(t->type.func->arg[i],n);
      if (!(*n)) break;
    }
    return st;
  } else {
    return (tree *) NULL;
  }
}

tree *get_function_subtree(
	tree *t,
	int n
	)
{
  int	m;

  m = n;
  return gfs(t,&m);
}

tree *gfs(
	tree *t,
	int *n
	)
{
  tree	*st;
  int	i;

  if (!(*n)) return(t);
  for (i=0; i<function_arity(t); i++) {
    if (t->type.func->arg[i]->nodetype == FUNCTION) {
      --(*n);
      st = gfs(t->type.func->arg[i],n);
      if (!(*n)) break;
    }
  }
  return st;
}

tree **pointer_to_subtree(
	tree **pointer,
	tree *subt
	)
{
  int	i;
  tree	**pos;

  if (*pointer == subt) {
    return pointer;
  }

  for (i=0; i<function_arity(*pointer); i++) {
    if ((*pointer)->type.func->arg[i]->nodetype == FUNCTION) {
      pos = pointer_to_subtree(&((*pointer)->type.func->arg[i]), subt);
      if ((pos != NULL) && ((*pos) == subt)) return pos;
    } else {
      if ((*pointer)->type.func->arg[i] == subt)
	return &((*pointer)->type.func->arg[i]);
    }
  }
  return ((tree **)NULL); /* make lint happy */
}
