/*
SGPC: Simple Genetic Programming in C
(c) 1993 by Walter Alden Tackett and Aviram Carmi

 This code and documentation is copyrighted and is not in the public domain.
 All rights reserved.

 - This notice may not be removed or altered.

 - You may not try to make money by distributing the package or by using the
   process that the code creates.

 - You may not distribute modified versions without clearly documenting your
   changes and notifying the principal author.

 - The origin of this software must not be misrepresented, either by
   explicit claim or by omission.  Since few users ever read sources,
   credits must appear in the documentation.

 - Altered versions must be plainly marked as such, and must not be
   misrepresented as being the original software.  Since few users ever read
   sources, credits must appear in the documentation.

 - The authors are not responsible for the consequences of use of this
   software, no matter how awful, even if they arise from flaws in it.

If you make changes to the code, or have suggestions for changes,
let us know!  (gpc@ipld01.hac.com)
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "magic.h"
#include "gpc.h"
#include "random.h"
#include "proto.h"
#include "globals.h"



void startup(int argc, char **argv, int *numpops, int *numgens,
             int *demes, int *demerows, int *demecols,
             int *start_gen, FILE **ckpt_file, pop_struct **pop,
             pop_struct    ****grid)
{
  int p;
  int popnum;
  int i;
  char	buf[132], val[132];

/* newly added 1994 3 12 */

  

  node_list = NULL;   // Not sure of what this does, it is used on creating trees


  gaussian_noise_toggle = 1;

  make_function_table(*numpops, *pop);
  make_terminal_table(*numpops, *pop);
  allocate_populations(*numpops,*pop);
  if (*demes) {
    setup_deme_grid(*numpops,*demerows,*demecols,*pop,*grid);
  }
  /*   define_fitness_cases(*numpops,*numgens,*pop); */
  if(*start_gen) {

    (*start_gen)++; /* restart from next gen after crash */
    for (p = 0; p < *numpops; p++) {
      popnum = -1;
      fscanf(*ckpt_file,"%s = %s", buf, val);
      if (strcmp("_START_OF_POPULATION_",buf)) {
	fprintf(stderr,"Error, Should have seen _START_OF_POPULATION_ = %d\n",
		p);
	exit(1);
      }
      sscanf(val,"%d",&popnum);
      if (popnum != p) {
	fprintf(stderr,"Error reading ckpt file: Population mismatch %d %d\n",
		p,popnum);
	exit(1);
      }
      printf("Reading population %d from checkpoint file\n", popnum);
      for (i = 0; i < (*pop)[p].population_size; i++) {
	(*pop)[p].population[i] = read_tree(*pop, p, *ckpt_file);
      }
    }

    for (p = 0; p < *numpops; p++) {

      for (i = 0; i < (*pop)[p].population_size; i++) {
	(*pop)[p].fitness_sort_index[i] = i;
      }

      zero_fitness_of_populations(*numpops,*pop,p);
      evaluate_fitness_of_populations(*numpops,*numgens,*pop,p);
      if ((*pop)[p].parsimony_factor > 0.0) add_parsimony_to_fitness(*pop,p);

      normalize_fitness_of_population(*pop,p);

      sort_population_by_fitness(*pop,p);

      fscanf(*ckpt_file,"%s = %s", buf, val);
      if (!strcmp("best_of_run_fitness",buf)) {
	sscanf(val,"%f",&((*pop)[p].best_of_run_fitness));
      }
      else {
	fprintf(stderr,"Error reading checkpoint file.\n");
      }
      fscanf(*ckpt_file,"%s = %s", buf, val);
      if (!strcmp("best_of_run_gen",buf)) {
	sscanf(val,"%d",&((*pop)[p].best_of_run_gen));
      }
      else {
	fprintf(stderr,"Error reading checkpoint file.\n");
      }
      (*pop)[p].best_of_run = read_tree(*pop, p, *ckpt_file);

    }

    fclose(*ckpt_file);
  }
}



/* this is not used anymore in the library
VOID run_gp_system(
  int		numpops,
  int  		numgens,
  int		start_gen,
  pop_struct 	*pop,
  int		demes,
  pop_struct	***grid,
  int		demerows,
  int		demecols
  )
{
    FILE *cf1,*cf2;
    char  ckptfile[256];


  long clocktime, clock P_((void));
  clock();


  write_header("dat");
  write_header("cas");

  generations(numpops,numgens,start_gen,pop,demes,grid,demerows,demecols);


  write_tailer("dat");
  write_tailer("cas");


  // NOTE *** clock() rolls over after about 36 minutes *
  clocktime = clock();
  printf("\ntime= %f seconds\n", ((double)clocktime)/1.0e6);


  sprintf(ckptfile,"%s_%ld_%d.ckpt","dat",42,42); //gethostid(),getpid());
  cf1 = fopen(ckptfile,"a");
  sprintf(ckptfile,"%s_%ld_%d.ckpt","cas",42,42);//gethostid(),getpid());
  cf2 = fopen(ckptfile,"a");
  fprintf(cf1,"\ntime= %f seconds\n", ((double)clocktime)/1.0e6);
  fprintf(cf1,"\ntime= %f seconds\n", ((double)clocktime)/1.0e6);
  fclose(cf1);
  fclose(cf2);



}

*/
