/*
SGPC: Simple Genetic Programming in C
(c) 1993 by Walter Alden Tackett and Aviram Carmi

 This code and documentation is copyrighted and is not in the public domain.
 All rights reserved.

 - This notice may not be removed or altered.

 - You may not try to make money by distributing the package or by using the
   process that the code creates.

 - You may not distribute modified versions without clearly documenting your
   changes and notifying the principal author.

 - The origin of this software must not be misrepresented, either by
   explicit claim or by omission.  Since few users ever read sources,
   credits must appear in the documentation.

 - Altered versions must be plainly marked as such, and must not be
   misrepresented as being the original software.  Since few users ever read
   sources, credits must appear in the documentation.

 - The authors are not responsible for the consequences of use of this
   software, no matter how awful, even if they arise from flaws in it.

If you make changes to the code, or have suggestions for changes,
let us know!  (gpc@ipld01.hac.com)
*/

#ifndef _GPC_H_
#define _GPC_H_

#include "magic.h" // Kludge, remove later



#define P_(s) s

/* needed constants */
#define VOID void
#define TRUE 1
#define FALSE 0

/* system symbolic names */
#define FUNCTION 0
#define TERMINAL 1

#define FULL   2
#define GROW   3
#define RAMPED 4

#define TOURNAMENT 5
#define OVERSELECT 6
#define FITNESSPROP 7

#define type_from_tree(t) ((t)->type.func->node_pointer->type)
#define	no_coeff_from_tree(t) ((type_from_tree(t) > 0) ? node_functions[type_from_tree(t)].arity1 : node_functions[- (type_from_tree(t))].arity2)
#define	no_coeff_from_type(typ) ((typ) > 0 ? node_functions[typ].arity1 : node_functions[-(typ)].arity2)
#define	function_printname(t) pop[(t)->pop].function_table[(t)->id].printname
#define	function_code_from_tree(t) ((type_from_tree(t) > 0) ? node_functions[type_from_tree(t)].code1 : node_functions[- type_from_tree(t)].code2)
#define	function_code_from_type(typ) ((typ > 0) ? node_functions[typ].code1 : node_functions[- (typ)].code2)
#define	terminal_printname(t) POP[(t)->pop].terminal_table[(t)->id].printname


typedef struct tnode {
  int		nodetype;
  int		id;
  int		pop;
  struct tnode  *parent;
  union {
    struct fdata	*func;
    struct tdata	*term;
  } type;
} tree;

typedef struct fdata {
  tree		**arg;
  struct nonterminal_node *node_pointer;
} function;

typedef struct fte {
  int		arity1; ///< number of parameter of P_ ??
  int		arity2; ///< number of parameter of P_ ??
  char		printname[15]; ///< name of function ??

  GENERIC *(*code1) P_((GENERIC *));
  GENERIC *(*code2) P_((GENERIC *));
} function_table_entry;


typedef struct tte {
  GENERIC	val;
  char		printname[15];
  GENERIC	(*constant_generator) P_(( ));
} terminal_table_entry;

typedef struct pop_struct{
  int			population_size;
  int			steady_state;
  int			my_row, my_col;
  char			*load_from_file;
  char			*format;
  char			*ckpt_format;
  int			max_depth_for_new_trees;
  int			max_depth_after_crossover;
  int			max_mutant_depth;
  int			grow_method;
  int			selection_method;
  int                   elite_size;
  int			tournament_K;
  int			*tournament_index;
  float			deme_search_radius_sigma;
  float			crossover_func_pt_fraction;
  float			crossover_any_pt_fraction;
  float			fitness_prop_repro_fraction;
  float			parsimony_factor;
  float			*mse;
  float			*mdl;
  float			*standardized_fitness;
  float			*adjusted_fitness;
  float			*normalized_fitness;
  int			*fitness_sort_index;
/* newly added 1994 3 12 */
  int			*parent1;
  int			*parent2;
  int			*cross_funct;
  int			*cross_any;
  int			*mut_pos;
  tree			**population;
  tree			**new_population;
  tree			*best_of_generation;
  tree			*best_of_run;
  int			best_of_run_gen;
  float			best_of_gen_fitness;
  float			best_of_run_fitness;
  float			best_of_gen_mse;
  float			best_of_run_mse;
  float			best_of_gen_mdl;
  float			best_of_run_mdl;
  function_table_entry 	*function_table; /* terminal_table has an extra entry:
					    terminal_table[terminal_table_size]
					    is the printf format string for use
					    with reading/printing of constants
					    */
  terminal_table_entry	*terminal_table;
  int			function_table_size;
  int			terminal_table_size;
  int                   elite_dupe;
  int                   crossover_dupe;
} pop_struct;

struct no_tag {
  int popno;
  int index;
  struct no_tag *next;
};

struct nonterminal_node {
/*   int name; */
  int type;
  double *coeff; ///< coefficients of GMDH
  double *xlist; ///< required to fit GMDH for objective function
  double mse; ///< mean square error
  struct no_tag *pop_pointer; ///< ????
  struct nonterminal_node *next; ///< next node of tree
};


#endif

/* Why is this after the initial endif? Shoulw be
 * restated everytime???? 01/09 Claus
 */

/* node-funs */
//#define MAX_NODE_FUNS 10
#define MAX_NODE_FUNS 20

#define DETAIL_DAT 0 /* detail-flag for sga.dat */
#define DETAIL_CAS 0 /* detail-flag for sga.cas */

#define NO_TYPE 99999 /* for node_type */







