/* Initialization functions for SGPC */
/* Created on 01/09 */
/* By Claus */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
//#include <dos.h>

#include "gpc.h"
#include "proto.h"
#include "gmdh.h"
#include "magic.h"
#include "problem.h"
#include "getparams.h"
#include "globals.h"
#include "treeio.h"

#include "init.h"

#include "prob.h"
#include "clindef.h"

/* This function scans for and eliminate duplicate
   trees from the population. Duplicate trees are
   mutated. It returns the number of duplicates
   found.
   This should be used after sorting the population
   by fitness, because it uses the fact that 
   identical trees will have identical fitnesses.
   Added by Claus 06/09/03
*/

void run_one_generation(int start_gen, int g, pop_struct *pop,pop_struct ***grid)
{
  int	p, i, j;

  clock_t t_start, t_end;
  t_start = clock();

  for (p = 0; p<P.numpops; p++)
  {
    for (i=0; i<pop[p].population_size; i++)
    {
      pop[p].parent1[i] = -1;
      pop[p].parent2[i] = -1;
      pop[p].cross_funct[i] = -1;
      pop[p].cross_any[i] = -1;
      pop[p].mut_pos[i] = -1;
    }
  }

  if (g) // not first generation
  {
    for (p = 0; p<P.numpops; p++)
    {
      free_tree(pop[p].best_of_generation);
      breed_new_population(pop,p,0,0,0,pop[p].steady_state);
      if (!(pop[p].steady_state))
      {
        free_population(pop,p);
        load_new_population(pop,p);
      }
    }
  }
  else // !g -> first generation
  {
    /* generation 0 is random */
    initialize_populations(P.numpops,pop);
    for (p = 0; p<P.numpops; p++)
    {
      if (pop[p].steady_state)
      {
	evaluate_fitness_of_populations(P.numpops,P.numgens,pop,p);
	if (pop[p].parsimony_factor > 0.0) add_parsimony_to_fitness(pop,p);
      }
    }
  } // finish running the population for this generation.

  /* NOTE that we breed ALL populations before we evaluate their fitness
     (except in the case of steady-state).  This is done to support
     co-evolution where there are multiple interacting populations
     who are co-evaluated at each generation */

  for (p = 0; p<P.numpops; p++)
  {
    if (!(pop[p].steady_state))
    {
      zero_fitness_of_populations(P.numpops,pop,p);
      evaluate_fitness_of_populations(P.numpops,P.numgens,pop,p);
      if (pop[p].parsimony_factor > 0.0) add_parsimony_to_fitness(pop,p);
    }

    normalize_fitness_of_population(pop,p);

    /*      sort_population_by_fitness(pop,p); */
    if (P.eval_by_mdl)
      sort_population_by_mdl(pop,p);

	
    else
      sort_population_by_mse(pop,p);
         

    pop[p].best_of_gen_mse =
      pop[p].mse[pop[p].fitness_sort_index[0]];
    pop[p].best_of_gen_mdl =
      pop[p].mdl[pop[p].fitness_sort_index[0]];
    pop[p].best_of_generation =
      copy_tree(pop[p].population[pop[p].fitness_sort_index[0]]);

    pop[p].best_of_run_mdl = pop[p].best_of_gen_mdl ;
    pop[p].best_of_run_mse = pop[p].best_of_gen_mse ;
    free((char *)pop[p].best_of_run);
    pop[p].best_of_run = copy_tree(pop[p].best_of_generation);
    pop[p].best_of_run_gen = g;


  } // num of pops

  gc_nodes(pop,P.numpops);

  t_end = clock();
  //  P.runtime += (double)(t_end - t_start) / (double)CLK_TCK;

}




