
#include <stdio.h>

#include "magic.h"
#include "gmdh.h"
#include "problem.h"
#include "getparams.h"
#include "output.h"
#include "globals.h"
#include "treeio.h"

/* Returns the value calculated by tree t for index i on data */
double test_value(tree *t, int i)
{
  double args[2], *ans, tmp;
  struct nonterminal_node *np;
  double *(*fn)(double *args);

  if (t->nodetype== TERMINAL)
    return (P.getTerminal(i,1+t->id));

  np = t->type.func->node_pointer;

  args[0]=test_value(t->type.func->arg[0],i);
  args[1]=test_value(t->type.func->arg[1],i);

  fn = function_code_from_type(np->type);
  ans = (*fn)(args);

  tmp = 0.0;

  for(int j=0; j<no_coeff_from_type(np->type);j++)
    tmp=tmp+(np->coeff[j])*ans[j];

  deallocvct(ans);
  return(tmp);
}

double test_value2(tree *t, int i)
{
  double args[2], *ans, tmp;
  struct nonterminal_node *np;
  double *(*fn)(double *args);

  if (t->nodetype == TERMINAL)
  {
    if ((i - (1+t->id) < 0)||
        (i - (1+t->id) >= P.data_max))
    {
      return 0;
    }
    else
    {
      if (Result.output2[i - (1+t->id)] > P.overlimit) return P.overlimit;
      if (Result.output2[i - (1+t->id)] < -1*P.overlimit) return -1*P.overlimit;
      return (Result.output2[i - (1+t->id)]);
    }
  }

  np = t->type.func->node_pointer;

  args[0]=test_value2(t->type.func->arg[0],i);
  args[1]=test_value2(t->type.func->arg[1],i);

  fn = function_code_from_type(np->type);
  ans = (*fn)(args);

  tmp = 0.0;

  for(int j=0; j<no_coeff_from_type(np->type);j++)
      tmp=tmp+(np->coeff[j])*ans[j];

  deallocvct(ans);
  
  return(tmp);
}

output::output()
{
}

void output::update_populations()
{
 double sum_mse, sum_mdl;

  // get the value for best and average MSE/MDL, and their index
  for (int i=0; i<P.numpops; i++)
  {
    best_mse[i] = (double) POP[i].best_of_gen_mse;
    best_mdl[i] = (double) POP[i].best_of_gen_mdl;

    sum_mse = 0;
    sum_mdl = 0;

    for (int j=0; j< POP[i].population_size; j++)
    {
      sum_mse+= (double) POP[i].mse[j];
      if ((double) POP[i].mse[j]==best_mse[i])
        best_ind_mse[i] = j;

      sum_mdl+= (double) POP[i].mdl[j];
      if ((double) POP[i].mdl[j]==best_mdl[i])
        best_ind_mdl[i] = j;

    }
    avg_mse[i] = sum_mse / POP[i].population_size;
    avg_mdl[i] = sum_mdl / POP[i].population_size;
  }
}

void output::set_value(int pop, int index)
{

  pop_number = index;
  ind_number = pop;

  mse = (double) POP[pop].mse[index];
  mdl = (double) POP[pop].mdl[index];
  tree[0]='\0';
  sprint_tree_a(POP[pop].population[index], POP[pop].ckpt_format, tree, 0);

}

void output::set_pred(int pop, int index)
{
  for (int i = 0; i < P.data_train_max; i++)
  {
    output2[i] = (double) POP[pop].population[index]->type.func->node_pointer->xlist[i];
  }
  for (int i = P.data_train_max; i < P.data_max; i++)
  {
    output2[i] = test_value(POP[pop].population[index], i);
  }
}

void output::set_pred2(int pop, int index)
{
  for (int i = 0; i < P.data_train_max; i++)
  {
    output2[i] = (double) POP[pop].population[index]->type.func->node_pointer->xlist[i];
  }
  for (int i = P.data_train_max; i < P.data_max; i++)
  {
    output2[i] = test_value2(POP[pop].population[index], i);
  }
}

void output::set_best_value(int pop)
{
  pop_number = -1;
  ind_number = pop;

  mse = (double) POP[pop].best_of_gen_mse;
  mdl = (double) POP[pop].best_of_gen_mdl;
  tree[0]='\0';
  sprint_tree_a(POP[pop].best_of_generation, POP[pop].ckpt_format, tree, 0);

}

void output::set_best_pred(int pop)
{
  for (int i = 0; i < P.data_train_max; i++)
  {
    output2[i] = (double) POP[pop].best_of_generation->type.func->node_pointer->xlist[i];
  }
  for (int i = P.data_train_max; i < P.data_max; i++)
  {
    output2[i] = test_value(POP[pop].best_of_generation, i);
  }
}

void output::set_best_pred2(int pop)
{

  //drawing REAL training curve
  for (int i = 0; i < POP[pop].terminal_table_size; i++)
  {
    output2[i] = P.data[i];
  }
  for (int i = POP[pop].terminal_table_size; i < P.data_train_max; i++)
  {
    // was 2
    output2[i] = test_value(POP[pop].best_of_generation, i);
  }

  // old function
  for (int i = P.data_train_max; i < P.data_train_max+P.data_gap; i++)
  {
    output2[i] = P.data[i];
  }
  for (int i = P.data_train_max+P.data_gap; i < P.data_max; i++)
  {
    // was 2
    output2[i] = test_value(POP[pop].best_of_generation, i);
  }
}


int sign(double x)
{
  if (x > 0)
    return 1;
  if (x < 0)
    return -1;
  return 0;
}


/**
 * \fn double hit_rate()
 * \brief Return a value that states the percentage of "correct" predictions based on upwards an downwards tendencies
 *
 * \return hit rate of predictions
 */
double output::hit_rate()
{
  double sum = 0;
  int n = 0;
  for (int i = P.data_train_max+P.data_gap; i < P.data_max; i++)
    {
      if (sign(output2[i]-output2[i-1])==sign(P.data[i]-P.data[i-1]))
	sum++;
      n++;
    }
  return (sum/n);
}
