/*
SGPC: Simple Genetic Programming in C
(c) 1993 by Walter Alden Tackett and Aviram Carmi
 
 This code and documentation is copyrighted and is not in the public domain.
 All rights reserved. 
 
 - This notice may not be removed or altered.
 
 - You may not try to make money by distributing the package or by using the
   process that the code creates.
 
 - You may not distribute modified versions without clearly documenting your
   changes and notifying the principal author.
 
 - The origin of this software must not be misrepresented, either by
   explicit claim or by omission.  Since few users ever read sources,
   credits must appear in the documentation.
 
 - Altered versions must be plainly marked as such, and must not be
   misrepresented as being the original software.  Since few users ever read
   sources, credits must appear in the documentation.
 
 - The authors are not responsible for the consequences of use of this 
   software, no matter how awful, even if they arise from flaws in it.
 
If you make changes to the code, or have suggestions for changes,
let us know!  (gpc@ipld01.hac.com)
*/



#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#ifdef MACOSX
#include <limits.h>
#include <float.h>
#define MAXFLOAT FLT_MAX
#else
#include <values.h>
#endif
#include <string.h>

#include "gpc.h"
#include "proto.h"

#include "magic.h"
#include "problem.h"
#include "getparams.h"
#include "globals.h"

#include "treeio.h"

VOID allocate_populations(
  int		numpops,
  pop_struct	*pop
  )
{
  int	p;

  for (p=0; p<numpops; p++) {
    pop[p].mse =
      (float *) malloc(pop[p].population_size*sizeof(float));
    pop[p].mdl =
      (float *) malloc(pop[p].population_size*sizeof(float));
    pop[p].standardized_fitness =
      (float *) malloc(pop[p].population_size*sizeof(float));
    pop[p].adjusted_fitness = (float *) malloc(pop[p].population_size*sizeof(float));
    pop[p].normalized_fitness =
      (float *) malloc(pop[p].population_size*sizeof(float));
    pop[p].fitness_sort_index = (int *) malloc(pop[p].population_size*sizeof(int));
    pop[p].tournament_index = (int *) malloc(pop[p].tournament_K*sizeof(int));
    pop[p].population = (tree **) malloc(pop[p].population_size*sizeof(tree *));
    pop[p].new_population = (tree **) malloc(pop[p].population_size*sizeof(tree *));
    pop[p].best_of_generation = (tree *) NULL; 
    pop[p].best_of_run = (tree *) NULL;
    pop[p].best_of_gen_fitness = (MAXFLOAT);
    pop[p].best_of_run_fitness = (MAXFLOAT);
/* newly added  1994 3 12 */
    pop[p].parent1 = (int *) malloc(pop[p].population_size*sizeof(int));
    pop[p].parent2 = (int *) malloc(pop[p].population_size*sizeof(int));
    pop[p].cross_funct = (int *) malloc(pop[p].population_size*sizeof(int));
    pop[p].cross_any = (int *) malloc(pop[p].population_size*sizeof(int));
    pop[p].mut_pos = (int *) malloc(pop[p].population_size*sizeof(int));
  }
}


VOID setup_deme_grid(
  int 		numpops,
  int		demerows,
  int		demecols,
  pop_struct 	*pop,
  pop_struct    ***grid
  )
{
  int	r, c;
  int	i, p, nperdeme;

  for (p=0;p<numpops;p++) {
    nperdeme = (pop[p].population_size/(demerows*demecols));
    for (r=0,i=0; r<demerows;r++) {
      for (c=0;c<demecols;c++,i+=nperdeme) {
	grid[r][c][p].my_row = r;
	grid[r][c][p].my_col = c;
	grid[r][c][p].population_size = nperdeme;
	grid[r][c][p].steady_state = pop[p].steady_state;
	grid[r][c][p].load_from_file = pop[p].load_from_file;
	grid[r][c][p].max_depth_for_new_trees =
	  pop[p].max_depth_for_new_trees;
	grid[r][c][p].max_depth_after_crossover =
	  pop[p].max_depth_after_crossover;
	grid[r][c][p].max_mutant_depth = pop[p].max_mutant_depth;
	grid[r][c][p].grow_method = pop[p].grow_method;
	grid[r][c][p].selection_method = pop[p].selection_method;
	grid[r][c][p].tournament_K = pop[p].tournament_K;
	grid[r][c][p].deme_search_radius_sigma =
	  pop[p].deme_search_radius_sigma;
	grid[r][c][p].tournament_index =
	  (int *) malloc(pop[p].tournament_K*sizeof(int));
	grid[r][c][p].crossover_func_pt_fraction =
	  pop[p].crossover_func_pt_fraction;
	grid[r][c][p].crossover_any_pt_fraction =
	  pop[p].crossover_any_pt_fraction;
	grid[r][c][p].fitness_prop_repro_fraction =
	  pop[p].fitness_prop_repro_fraction;
	grid[r][c][p].parsimony_factor = pop[p].parsimony_factor;
	grid[r][c][p].standardized_fitness =
	  &(pop[p].standardized_fitness[i]);
	grid[r][c][p].adjusted_fitness =
	  &(pop[p].adjusted_fitness[i]);
	grid[r][c][p].normalized_fitness =
	  &(pop[p].normalized_fitness[i]);
	grid[r][c][p].fitness_sort_index =
	  &(pop[p].fitness_sort_index[i]);
	grid[r][c][p].population = &(pop[p].population[i]);
	grid[r][c][p].new_population =
	  &(pop[p].new_population[i]);
	grid[r][c][p].best_of_generation = (tree *) NULL;
	grid[r][c][p].best_of_run = (tree *) NULL;
	grid[r][c][p].best_of_run_gen = -1;
	grid[r][c][p].best_of_gen_fitness = -1.0;
	grid[r][c][p].best_of_run_fitness = -1.0;
	grid[r][c][p].function_table = pop[p].function_table; 
	grid[r][c][p].terminal_table = pop[p].terminal_table;
	grid[r][c][p].function_table_size = pop[p].function_table_size;
	grid[r][c][p].terminal_table_size = pop[p].terminal_table_size;
      }
    }
  }
}

VOID initialize_populations(
  int		numpops,
  pop_struct	*pop
  )
{
  int	p;
  int	i;
  int	min_depth_for_new_trees = 1;
  int 	full_cycle;
  int 	grow;
  int	size;
  FILE	*f;
  tree	*temp;

  for (p=0; p<numpops; p++) {
    full_cycle = 0;
    for (i=0; i<pop[p].population_size; i++) {
      switch (pop[p].grow_method) {
      case FULL:
	size = pop[p].max_depth_for_new_trees;
	grow = 1;
	break;
      case GROW:
	size = pop[p].max_depth_for_new_trees;
	grow = 0;
	break;
      case RAMPED:
	size = (min_depth_for_new_trees +
		(i % (pop[p].max_depth_for_new_trees-min_depth_for_new_trees)));
	if (pop[p].max_depth_for_new_trees != min_depth_for_new_trees)
	  if (!(i % (pop[p].max_depth_for_new_trees-min_depth_for_new_trees)))
	    full_cycle = (!full_cycle);
	grow = (full_cycle);
	break;
      default:
	fprintf(stderr,"Error in initialize_populations(): Method %d must be %d %d or %d\n",
		pop[p].grow_method, FULL, GROW, RAMPED);
      }
      pop[p].population[i] = create_random_tree(pop, p, size, 1, grow, NULL);
    }
    /* replace the first members of population with those to be read
       from a file, if filename is not null */
    if (pop[p].load_from_file[0] != '\0') {
      f = fopen(pop[p].load_from_file,"r");
      for (i=0; (((int)(temp=read_tree(pop,p,f))) != EOF) &&
	   (i<pop[p].population_size); i++) {
	free_tree(pop[p].population[i]);
	pop[p].population[i] = temp;
      }
    }
  }
}

VOID breed_new_population(
  pop_struct	*pop,
  int 		p,
  int 		demes,
  int 		nrows,
  int 		ncols,
  int		steady_state
  )
{
  int	i, j, incr;
  float fraction = 0.0;
  tree	*parent1, *parent2, *offspring1, *offspring2, *pptr;
  int	worst_index1, worst_index2, best_index1, best_index2;
  int   jcross1, jcross2, mut_pos;

  if (steady_state && ((pop[p].selection_method == OVERSELECT) ||
		      (pop[p].selection_method == FITNESSPROP))) {
    printf("SORRY: steady_state population requires either DEMES or \n \
            selection_method = TOURNAMENT");
    exit (1);
  }

  /* Elite strategy, copy the best X individuals to the next generation */
  for (i = 0,j = 0; i < pop[p].elite_size; j++, i++)
  {
    while (i > 0 &&
	   pop[p].elite_dupe == 0 &&
	   j < pop[p].population_size - pop[p].elite_size &&
	   tree_cmp(pop[p].new_population[i-1], 
		    pop[p].population[pop[p].fitness_sort_index[j]]) == 0)
      {
	j++;
      }

    pop[p].new_population[i] = copy_tree(pop[p].population[pop[p].fitness_sort_index[j]]);
    

  }


  /* End Elite strategy */


  for(i = pop[p].elite_size, fraction = random_float(1.0);//add12/03
      i < pop[p].population_size;
      i += incr, fraction = random_float(1.0))
  {
    parent1 = find_tree(pop,p,demes,nrows,ncols,&worst_index1,&best_index1);

    while (steady_state && (POP[p].population[worst_index1] == parent1))
      parent1 = find_tree(pop,p,demes,nrows,ncols,&worst_index1,&best_index1);

    if ((i < (pop[p].population_size-1)) &&
	(fraction <
	 (pop[p].crossover_func_pt_fraction+pop[p].crossover_any_pt_fraction)))
    {
      parent2 = find_tree(pop,p,demes,nrows,ncols,&worst_index2,&best_index2);
      while (steady_state && ((worst_index1 == worst_index2) ||
			      (POP[p].population[worst_index2] == parent2) ||
			      (POP[p].population[worst_index2] == parent1) ||
			      (POP[p].population[worst_index1] == parent2)))
	parent2 = find_tree(pop,p,demes,nrows,ncols,&worst_index2,&best_index2);

      if (fraction < pop[p].crossover_func_pt_fraction)
      {
	crossover_at_func_pt(pop, parent1, parent2, &offspring1, &offspring2,
			     &jcross1, &jcross2);
      }
      else
      {
	crossover_at_any_pt(pop, parent1, parent2, &offspring1, &offspring2,
			     &jcross1, &jcross2);
      }

      /* If crossover_dupe parameter is 0, mutate offspring 
       * which are identical to parents.
       */
      if (pop[p].crossover_dupe == 0 &&
	  (!(tree_cmp(offspring1,parent1) && tree_cmp(offspring1,parent2))))
	{
	  offspring1 = mutate(pop,offspring1, &mut_pos);
	}
      if (pop[p].crossover_dupe == 0 &&
	  (!(tree_cmp(offspring2,parent1) && tree_cmp(offspring2,parent2))))
	{ 
	  offspring2 = mutate(pop,offspring2, &mut_pos);
	} 

      if (steady_state) 
	{
	  free_tree(POP[p].population[worst_index1]);
	  free_tree(POP[p].population[worst_index2]);
	  POP[p].population[worst_index1] = offspring1;
	  POP[p].population[worst_index2] = offspring2;
	  /* newly added 1994 3 12 */
	  POP[p].parent1[worst_index1] = best_index1;
	  POP[p].parent2[worst_index1] = best_index2;
	  POP[p].parent1[worst_index2] = best_index2;
	  POP[p].parent2[worst_index2] = best_index1;
	  if (fraction < pop[p].crossover_func_pt_fraction) {
	    POP[p].cross_funct[worst_index1] = jcross1;
	    POP[p].cross_funct[worst_index2] = jcross2;
	    POP[p].mut_pos[worst_index1] = -1;
	    POP[p].mut_pos[worst_index2] = -1;
	    POP[p].cross_any[worst_index1] = -1;
	    POP[p].cross_any[worst_index2] = -1;
	  } else {
	    POP[p].cross_any[worst_index1] = jcross1;
	    POP[p].cross_any[worst_index2] = jcross2;
	    POP[p].mut_pos[worst_index1] = -1;
	    POP[p].mut_pos[worst_index2] = -1;
	    POP[p].cross_funct[worst_index1] = -1;
	    POP[p].cross_funct[worst_index2] = -1;
	  }

	  POP[p].standardized_fitness[worst_index1] =
	    evaluate_fitness_of_individual(POP,p,
					   POP[p].population[worst_index1],
					   worst_index1) +
	    ((float) count_crossover_pts(POP[p].population[worst_index1]))*
	    pop[p].parsimony_factor;
	  
	  POP[p].standardized_fitness[worst_index2] =
	    evaluate_fitness_of_individual(POP,p,
					   POP[p].population[worst_index2],
					   worst_index2) +
	    ((float) count_crossover_pts(POP[p].population[worst_index2]))*
	    pop[p].parsimony_factor;
	} 
      else 
	{
	  pop[p].new_population[i] = offspring1;
	  pop[p].new_population[i+1] = offspring2;
	  /* newly added 1994 3 12 */
	  pop[p].parent1[i] = best_index1;
	  pop[p].parent2[i] = best_index2;
	  pop[p].parent1[i+1] = best_index2;
	  pop[p].parent2[i+1] = best_index1;
	  if (fraction < pop[p].crossover_func_pt_fraction) {
	    pop[p].cross_funct[i] = jcross1;
	    pop[p].cross_funct[i+1] = jcross2;
	    pop[p].mut_pos[i] = -1;
	    pop[p].mut_pos[i+1] = -1;
	    pop[p].cross_any[i] = -1;
	    pop[p].cross_any[i+1] = -1;
	  } else {
	    pop[p].cross_any[i] = jcross1;
	    pop[p].cross_any[i+1] = jcross2;
	    pop[p].mut_pos[i] = -1;
	    pop[p].mut_pos[i+1] = -1;
	    pop[p].cross_funct[i] = -1;
	    pop[p].cross_funct[i+1] = -1;
	  }

	}
      incr = 2;
    }
    else if (fraction <
	     (pop[p].crossover_func_pt_fraction
	      + pop[p].crossover_any_pt_fraction
	      + pop[p].fitness_prop_repro_fraction)) {
      if (steady_state) {
#if DBSS == 1
	printf("performing straight copy of parent 1\n");
#endif
	free_tree(POP[p].population[worst_index1]);
	POP[p].population[worst_index1] = copy_tree(parent1);
	/* kind of a waste - you should really set fitess-prop-repro-frac
       	   to zero for the steady-state elitist model... */
	POP[p].standardized_fitness[worst_index1] =
	  evaluate_fitness_of_individual(POP,p,
					 POP[p].population[worst_index1],
					 worst_index1) +
	  ((float) count_crossover_pts(POP[p].population[worst_index1]))*
	  pop[p].parsimony_factor;
	
	/* newly added 1994 3 12 */
	POP[p].parent1[worst_index1] = best_index1;
	POP[p].parent2[worst_index1] = -1;
	POP[p].cross_funct[worst_index1] = -1;
	POP[p].mut_pos[worst_index1] = -1;
	POP[p].cross_any[worst_index1] = -1;
	
      } else {
	pop[p].new_population[i] = copy_tree(parent1);
	/* newly added 1994 3 12 */
	pop[p].parent1[i] = best_index1;
	pop[p].parent2[i] = -1;
	pop[p].cross_funct[i] = -1;
	pop[p].mut_pos[i] = -1;
	pop[p].cross_any[i] = -1;
	
      }
      incr = 1;
    } else {
      if (steady_state) {
#if DBSS == 1
	printf("performing mutation of parent 1\n");
#endif
	free_tree(POP[p].population[worst_index1]);
	POP[p].population[worst_index1] = mutate(pop,copy_tree(parent1),&mut_pos);
	POP[p].standardized_fitness[worst_index1] =
	  evaluate_fitness_of_individual(POP,p,
					 POP[p].population[worst_index1],
					 worst_index1) +
	  ((float) count_crossover_pts(POP[p].population[worst_index1]))*
	  pop[p].parsimony_factor;
/* newly added 1994 3 12 */
	POP[p].parent1[worst_index1] = best_index1;
	POP[p].parent2[worst_index1] = -1;
	POP[p].cross_funct[worst_index1] = -1;
	POP[p].mut_pos[worst_index1] = mut_pos;
	POP[p].cross_any[worst_index1] = -1;
	
      } else {
	pop[p].new_population[i] = mutate(pop,copy_tree(parent1), &mut_pos);
	/* newly added 1994 3 12 */
	pop[p].parent1[i] = best_index1;
	pop[p].parent2[i] = -1;
	pop[p].cross_funct[i] = -1;
	pop[p].mut_pos[i] = mut_pos;
	pop[p].cross_any[i] = -1;
	
      }
      incr = 1;
    }
  }
}

VOID free_population(
  pop_struct	*pop,
  int 	p
  )
{
  int	i;

  for (i=0; i<pop[p].population_size; i++) {
    free_tree(pop[p].population[i]);
  }
}

VOID load_new_population(
  pop_struct	*pop,
  int p
  )
{
  int	i;

  for (i=0; i<pop[p].population_size; i++) {
    pop[p].population[i] = pop[p].new_population[i];
  }
}


  
       
      
