/*
SGPC: Simple Genetic Programming in C
(c) 1993 by Walter Alden Tackett and Aviram Carmi
 
 This code and documentation is copyrighted and is not in the public domain.
 All rights reserved. 
 
 - This notice may not be removed or altered.
 
 - You may not try to make money by distributing the package or by using the
   process that the code creates.
 
 - You may not distribute modified versions without clearly documenting your
   changes and notifying the principal author.
 
 - The origin of this software must not be misrepresented, either by
   explicit claim or by omission.  Since few users ever read sources,
   credits must appear in the documentation.
 
 - Altered versions must be plainly marked as such, and must not be
   misrepresented as being the original software.  Since few users ever read
   sources, credits must appear in the documentation.
 
 - The authors are not responsible for the consequences of use of this 
   software, no matter how awful, even if they arise from flaws in it.
 
If you make changes to the code, or have suggestions for changes,
let us know!  (gpc@ipld01.hac.com)
*/



/*
 *
 * $Log: treegen.cpp,v $
 * Revision 1.3  2006/09/03 07:19:54  caranha
 * - Fixed clear_globals function so that it works. must be called after
 * running the program to prevent memory leaks for multiple runs/multiple
 * populations (globals.cpp/proto.h/main.cpp)
 *
 * - Removed some non-needed warning messages, replaced others to write to stderr
 * (gmdh.cpp)
 *
 * Revision 1.2  2006/09/02 22:33:04  yanase
 * add rename_tree_hold_type() in treegen.cpp/proto.h
 * to maintain the function node type after crossover and mutation.
 * replaced rename_tree by rename_tree_hold_type().
 * see also crossover.cpp and mutation.cpp.
 *
 * Revision 1.1.1.1  2006/08/31 08:56:26  yanase
 * This is the implementation of Inductive GP written in Adaptive Learning of Polynomial Networks.
 *
 * Revision 1.1.1.1  2006/08/21 09:52:44  caranha
 * Uploaded strogonoff Unix 02/2006 version to the server
 *
 * Revision 1.4  2006/01/31 05:49:01  caranha
 * *** empty log message ***
 *
 * Revision 1.3  2006/01/26 15:55:14  caranha
 * code cleanup, output() bug fixed (i think)
 *
 * Revision 1.2  2006/01/24 07:45:51  caranha
 * Nothing really important, just moving globals around
 *
 * Revision 1.1  2006/01/24 03:54:24  caranha
 * Added initial files - stroglib is the library, debug-app is a debug application for the library - should not be needed
 *
 * Revision 2.6  1993/04/22  07:39:12  gpc-avc
 * Removed old log messages
 *
 * Revision 2.5  1993/04/14  04:38:08  gpc-avc
 * Fixed side effect of read_tree() so it does not change random seed
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "gpc.h"
#include "proto.h"
#include "magic.h"
#include "globals.h"
#include "gmdh.h"

#define FRACTION_OF_NODE_CHOICE

tree *create_random_tree(
  pop_struct 	*pop,
  int 		p,
  int 		max_depth,
  int 		root_p,
  int 		full_p,
  tree          *parent
  )
{
  tree	*t;
  int	i, choice;
  struct nonterminal_node *new_node;

  if (max_depth <= 0)
  {
    t = create_tree_node(p, TERMINAL,
    random_int(pop[p].terminal_table_size+ALLOW_CONST), parent);
    return t;
  }
  else if (root_p || full_p)
  {
    choice = 0; /* one kind node */
    t = create_tree_node(p, FUNCTION, choice, parent);

    if ((t->type.func-> node_pointer =
	 (struct nonterminal_node *) malloc(sizeof(struct nonterminal_node))) == NULL) {
      perror(" at point 3 in init_function() <treegen.c>");
      exit(errno);
    }
    new_node = t->type.func-> node_pointer;
    new_node -> type = NO_TYPE;  /* initialize */
    new_node -> next = node_list;
    new_node -> coeff = NULL;
    new_node -> mse = -1.0;  /* initialize */
    new_node -> xlist = NULL;
    new_node -> pop_pointer = NULL;
    node_list = new_node;

    for (i=0; i<function_arity(t); i++) {
      t->type.func->arg[i] =
	create_random_tree(pop, p, (max_depth-1), 0, full_p, t);
    }
    return t;
  } else {
    choice =  random_int(pop[p].function_table_size +
			 pop[p].terminal_table_size+ALLOW_CONST);
#ifdef FRACTION_OF_NODE_CHOICE
    if(random_float(1.0) < Param.func_gen_fraction[p]) 
#else
    if (choice < pop[p].function_table_size) 
#endif
      {
	/*      while (!pop[p].function_table[choice].enabled)
		choice = random_int(pop[p].function_table_size); */
	choice = 0; /* one kind node */
	t = create_tree_node(p, FUNCTION, choice, parent);
	
	if ((t->type.func-> node_pointer =
	     (struct nonterminal_node *) malloc(sizeof(struct nonterminal_node))) == NULL) {
	  perror(" at point 3 in init_function() <treegen.c>");
	  exit(errno);
	}
	new_node = t->type.func-> node_pointer;
	new_node -> type = NO_TYPE;  /* initialize */
	new_node -> next = node_list;
	new_node -> coeff = NULL;
	new_node -> mse = -1.0;  /* initialize */
	new_node -> xlist = NULL;
	new_node -> pop_pointer = NULL;
	node_list = new_node;
	
	for (i=0; i<function_arity(t); i++) {
	  t->type.func->arg[i] =
	    create_random_tree(pop, p, (max_depth-1), 0, full_p, t);
	}
	return t;
      } else {
#ifdef FRACTION_OF_NODE_CHOICE
	choice =  random_int(pop[p].terminal_table_size+ALLOW_CONST);
	t = create_tree_node(p, TERMINAL, choice, parent);
#else
	t = create_tree_node(p, TERMINAL, choice-pop[p].function_table_size, parent);
#endif
	/* 
	 *(t->type.term->valptr) = (*(terminal_constant_generator(t)))(); 
	 */
	return t;
    }
  }
}


void rename_node(tree *t)
{
  struct nonterminal_node *new_node;

  if (t==NULL) return;

  if ((t->type.func-> node_pointer =
       (struct nonterminal_node *) malloc(sizeof(struct nonterminal_node))) == NULL) {
    perror(" at point 3 in init_function() <treegen.c>");
    exit(errno);
  }
  new_node = t->type.func-> node_pointer;
  new_node -> next = node_list;
  new_node -> type = NO_TYPE;  /* initialize */
  new_node -> coeff = NULL;
  new_node -> mse = -1.0;
  new_node -> xlist = NULL;
  new_node -> pop_pointer = NULL;
  node_list = new_node;
}

/**
 * \fn void rename_node_hold_type(tree *t)
 * \brief used only by rename_tree_hold_type: rename function node while holding the function type.
 * @author Toshihiko YANASE
 *
 * \param t tree
 */
void rename_node_hold_type(tree *t)
{
  struct nonterminal_node *new_node;

  if (t==NULL) return;

  int tmp = t->type.func->node_pointer->type;

  if ((t->type.func-> node_pointer =
       (struct nonterminal_node *) malloc(sizeof(struct nonterminal_node))) == NULL) {
    perror(" at point 3 in init_function() <treegen.c>");
    exit(errno);
  }
  new_node = t->type.func-> node_pointer;
  new_node -> next = node_list;
  new_node -> type = tmp;
  new_node -> coeff = NULL;
  new_node -> mse = -1.0;
  new_node -> xlist = NULL;
  new_node -> pop_pointer = NULL;
  node_list = new_node;


}

/**
 * \fn void rename_tree_hold_type(tree *t)
 * \brief rename function node while holding the function type.
 * @author Toshihiko YANASE
 *
 * \param t tree
 */
 void rename_tree_hold_type(tree *pos)
 {
  if (!(pos->parent == NULL)) rename_tree_hold_type(pos->parent);
  rename_node_hold_type(pos->parent);
 }


void rename_tree(tree *pos)
/* from the parent of pos */
{

  if (!(pos->parent == NULL)) rename_tree(pos->parent);
  rename_node(pos->parent);
}


tree  
*get_parent_aux(tree *root, tree *me, tree *parent)
{
  int i; 
  tree *left, *right, *tmp;

  if (root == me) return(parent);
  
  if (root->nodetype == TERMINAL) return(NULL);

  left=root->type.func->arg[0];
  tmp=get_parent_aux(left,me,root);
  if (tmp) return(tmp);

  right=root->type.func->arg[1];
  tmp=get_parent_aux(right,me,root);
  if (tmp) return(tmp);
  
  return(NULL);

}

tree *get_parent(
		 tree *root, tree *me)
{
  tree *left, *right, *tmp;

  if (root==me) return NULL;

  left=root->type.func->arg[0];
  tmp=get_parent_aux(left,me,root);
  if (tmp) return(tmp);

  right=root->type.func->arg[1];
  tmp=get_parent_aux(right,me,root);
  if (tmp) return(tmp);
  
  fprintf(stderr, "ERROR in parent get ... \n");
  return(NULL);
}
