/*
SGPC: Simple Genetic Programming in C
(c) 1993 by Walter Alden Tackett and Aviram Carmi

 This code and documentation is copyrighted and is not in the public domain.
 All rights reserved.

 - This notice may not be removed or altered.

 - You may not try to make money by distributing the package or by using the
   process that the code creates.

 - You may not distribute modified versions without clearly documenting your
   changes and notifying the principal author.

 - The origin of this software must not be misrepresented, either by
   explicit claim or by omission.  Since few users ever read sources,
   credits must appear in the documentation.

 - Altered versions must be plainly marked as such, and must not be
   misrepresented as being the original software.  Since few users ever read
   sources, credits must appear in the documentation.

 - The authors are not responsible for the consequences of use of this
   software, no matter how awful, even if they arise from flaws in it.

If you make changes to the code, or have suggestions for changes,
let us know!  (gpc@ipld01.hac.com)
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include "gpc.h"
#include "treeio.h"
#include "proto.h"
#include "magic.h"
#include "globals.h"


int read_terminal_set_values(pop_struct *pop,int p,FILE *f)
{
  int	i;

  for (i=0; i<pop[p].terminal_table_size; i++) {
    if (fscanf(f,pop[p].format, &(pop[p].terminal_table[i].val)) == EOF)
      return EOF;
  }
  return 1;
}

VOID load_terminal_set_values(pop_struct *pop,int p,GENERIC *v)
{
  int	i;

  for (i=0; i<pop[p].terminal_table_size; i++) {
    pop[p].terminal_table[i].val = v[i];
  }
}

#define TRACE 0
#ifdef ANSI_FUNC

tree *read_tree(
  pop_struct 	*pop,
  int 		p,
  FILE 		*f
  )
#else

tree *read_tree(
		pop_struct 	*pop,
		int		p,
		FILE		*f)
#endif
{
  int	i;
  tree 	*t, *parent;
  char	buf[80];
  
  if (fscanf(f,"%1s",buf) == EOF) return (tree *) EOF;
  if (buf[0] == '(') {

    if (TRACE) printf("(");

    fscanf(f,"%1s", buf);
    ungetc(buf[0],f);
    get_next_token(buf, f);

    if (TRACE) printf("%s ",buf);

    t = create_tree_node(p,FUNCTION, lookup_function_name(pop,p,buf), parent);
    //    t = create_tree_node(p,FUNCTION, lookup_function_name(pop,p,buf));
    for (i=0; i < function_arity(t); i++) {
      t->type.func->arg[i] = read_tree(pop,p,f);
    }

    fscanf(f,")");
    if (TRACE) printf(")\n");

  } else {
    ungetc(buf[0], f);
    get_next_token(buf, f);

    if (TRACE) printf("%s ",buf);

    t = create_tree_node(p,TERMINAL, lookup_terminal_name(pop,p,buf), parent);
/*    if (terminal_is_constant(t)) {
      sscanf(buf, pop[p].format, (t->type.term->valptr));
    } */
  }
  return(t);
}

VOID write_tree(
  pop_struct	*pop,
  tree 		*t,
  char		*format,
  FILE 		*f
  )
{
  pprint_tree(pop, t, format, f, 0);
}


VOID pprint_tree(
  pop_struct	*pop,
  tree 		*t,
  char		*format,
  FILE 		*f,
  int 		level
  )
{
  int	i;

  blanks(level,f);
  if (t->nodetype == FUNCTION) {
/*     fprintf(f,"(%s\n",function_printname(t)); */
    fprintf(f,"(NODE%d\n", t->type.func->node_pointer);
   // printf("Coeff of  NODE %d---------------------------- \n",t->type.func->node_pointer);//add1202
    for (i=0; i<no_coeff_from_type(t->type.func->node_pointer->type); i++) {
//	   printf("[%d]:%lf\n ",i,t->type.func->node_pointer->coeff[i]);//add11/30
    }
  //printf("over  NODE %d---------------------------------\n",t->type.func->node_pointer);

    for (i=0; i<function_arity(t); i++) {
      pprint_tree(pop, t->type.func->arg[i], format, f, level+INDENT);
      if (i == (function_arity(t)-1))
	fprintf(f,")");
      else
	fprintf(f,"\n");
    }
  } else if (t->nodetype == TERMINAL) {
 /* a constant */
/*    if (t->id == pop[t->pop].terminal_table_size) {
      fprintf(f, format, *(t->type.term->valptr));
    } else { */
      fprintf(f,"%s",terminal_printname(t));
/*    } */
  } else {
    fprintf(stderr,
	    "nodetype %d must be %d or %d in pprint_tree() <treeio.c>\n",
	    t->nodetype, FUNCTION, TERMINAL);
  }
  if (!level) fprintf(f,"\n");
}
//-----------------------------------------------------add12/03

VOID write_tree_a(
	      	pop_struct    *pop,
	      	tree          *t,
	      	char          *format,
	      	FILE          *f
	       )
{
      	pprint_tree_a(pop, t, format, f, 0);
}


VOID pprint_tree_a(
	      	pop_struct    *pop,
	      	tree          *t,
	      	char          *format,
	      	FILE          *f,
	      	int           level
	      	)
{
       	int   i;

	blanks(level,f);
      	if (t->nodetype == FUNCTION) {
	     /*fprintf(f,"(%s\n",function_printname(t)); */
	// fprintf(f,"(NODE%d\n", t->type.func->node_pointer);
	  fprintf(f,"(NODE%d    :",t->type.func->node_pointer);//add1202
		 for (i=0; i<no_coeff_from_type(t->type.func->node_pointer->type); i++) {
		fprintf(f,"[%d]%lf ",i,t->type.func->node_pointer->coeff[i]);//add11/30
		 }
		 fprintf(f,"\n");
		 for (i=0; i<function_arity(t); i++) {
		   	 pprint_tree_a(pop, t->type.func->arg[i], format, f, level+INDENT);
			 if (i == (function_arity(t)-1))
	                      fprintf(f,")");
		         else
		              fprintf(f,"\n");
		 }
  } else if (t->nodetype == TERMINAL) {
	   /* a constant */
	  /*    if (t->id == pop[t->pop].terminal_table_size) {
	   *          fprintf(f, format, *(t->type.term->valptr));
	   *              } else { */
	        fprintf(f,"%s",terminal_printname(t));
		/*    } */
		  } else {
	      fprintf(stderr,
                  "nodetype %d must be %d or %d in pprint_tree() <treeio.c>\n",
                  t->nodetype, FUNCTION, TERMINAL);
        }
  if (!level) fprintf(f,"\n");
}

VOID sprint_tree_a(tree *t, char *format, char *f,int level)
{
  int   i;
  char tmp[20];

  for (i=0; i < level; i++)
  {
    strcat(f," ");
  }

  if (t->nodetype == FUNCTION)
  {
    sprintf(tmp,"(NODE%d    :",t->type.func->node_pointer);//add1202
    strcat(f,tmp);

    sprintf(tmp, "%d  :",t->type.func->node_pointer->type);
    strcat(f,tmp);

    for (i=0; i<no_coeff_from_type(t->type.func->node_pointer->type); i++)
    {
      sprintf(tmp,"[%d]%lf ",i,t->type.func->node_pointer->coeff[i]);//add11/30
      strcat(f,tmp);
    }

    strcat(f,"\n");

    for (i=0; i<function_arity(t); i++)
    {
      sprint_tree_a(t->type.func->arg[i], format, f, level+INDENT);
      if (i == (function_arity(t)-1))
        strcat(f,")");
      else
	strcat(f,"\n");
    }
  }
  else if (t->nodetype == TERMINAL)
  {
    sprintf(tmp,"%s",terminal_printname(t));
    strcat(f,tmp);
  }
  else
  {
    fprintf(stderr,
            "nodetype %d must be %d or %d in pprint_tree() <treeio.c>\n",
            t->nodetype, FUNCTION, TERMINAL);
  }
  if (!level) strcat(f,"\n");
}

VOID sprint_tree_b(tree *t, char *format, char *f,int level)
{
  int   i;
  char tmp[20];

  for (i=0; i < level; i++)
  {
    strcat(f," ");
  }

  if (t->nodetype == FUNCTION)
  {
    sprintf(tmp, "(NODE T%2d:",t->type.func->node_pointer->type);
    strcat(f,tmp);

    strcat(f,"\n");

    for (i=0; i<function_arity(t); i++)
    {
      sprint_tree_b(t->type.func->arg[i], format, f, level+INDENT);
      if (i == (function_arity(t)-1))
        strcat(f,")");
      else
	strcat(f,"\n");
    }
  }
  else if (t->nodetype == TERMINAL)
  {
    sprintf(tmp,"%s",terminal_printname(t));
    strcat(f,tmp);
  }
  else
  {
    fprintf(stderr,
            "nodetype %d must be %d or %d in pprint_tree() <treeio.c>\n",
            t->nodetype, FUNCTION, TERMINAL);
  }
  if (!level) strcat(f,"\n");
}




//-------------------------------------------------------




int lookup_function_name(
  pop_struct 	*pop,
  int 		p,
  char 		*buf
  )
{
  int	i;

  for (i=0; i<pop[p].function_table_size; i++) {
    if (!strcmp(buf, pop[p].function_table[i].printname)) {
      return(i);
    }
  }

  fprintf(stderr,"%s not found in function_table[%d]\n",buf,p);
  exit(1);
  return(-1); /* make lint happy */
}


int lookup_terminal_name(
  pop_struct	*pop,
  int 		p,
  char 		*buf
  )
{
  int	i;

  for (i=0; i<pop[p].terminal_table_size; i++) {
    if (!strcmp(buf, pop[p].terminal_table[i].printname)) {
      return(i);
    }
  }
  /* Assume that buf is a constant value */
  return(pop[p].terminal_table_size);
}


VOID get_next_token(
	char *buf,
	FILE *f
	)
{
  int	ibuf;
  int	i = 0;

  while ((ibuf = getc(f)) != EOF) {
    if (isspace((char) ibuf)) {
      buf[i] = '\0';
      break;
    } else if (((char) ibuf) == ')') {
      ungetc((char) ibuf, f);
      buf[i] = '\0';
      break;
    } else {
      buf[i++] = (char) ibuf;
    }
  }
}

VOID blanks(
	int n,
	FILE *f
	)
{
  int	i;

  for (i=0; i<n; i++) fprintf(f," ");
}



